@@ -62,9 +62,12 @@ class AuthEnforcer(Fixture):
 
     def abort_or_redirect(self, page, message=""):
         """
-        return HTTP 403 if 'application/json' in HTTP_ACCEPT
-        else redirects to page"""
-        if re.search(REGEX_APPJSON, request.headers.get("accept", "")):
+        return HTTP 403 if 'application/json' in HTTP_ACCEPT and HTTP_JSON_REDIRECTS flag is not set in the request to 'on'
+        else redirects to page
+        """
+        if re.search(REGEX_APPJSON, request.headers.get("accept", "")) and (
+            request.headers.get("json-redirects", "") != "on"
+        ):
             abort(403)
         redirect_next = request.fullpath
         if request.query_string:
@@ -172,11 +175,11 @@ class Auth(Fixture):
         registration_requires_confirmation=True,
         registration_requires_approval=False,
         inject=True,
-        extra_fields=[],
+        extra_fields=None,
         login_expiration_time=3600,  # seconds
-        password_complexity={"entropy": 50},
+        password_complexity="default",
         block_previous_password_num=None,
-        allowed_actions=["all"],
+        allowed_actions=None,
         use_appname_in_redirects=None,
     ):
 
@@ -185,13 +188,16 @@ class Auth(Fixture):
             registration_requires_approval=registration_requires_approval,
             login_after_registration=False,
             login_expiration_time=login_expiration_time,  # seconds
-            password_complexity=password_complexity,
+            password_complexity={"entropy": 50}
+            if password_complexity == "default"
+            else password_complexity,
             block_previous_password_num=block_previous_password_num,
-            allowed_actions=allowed_actions,
+            allowed_actions=allowed_actions or ["all"],
             use_appname_in_redirects=use_appname_in_redirects,
             formstyle=FormStyleDefault,
             messages=copy.deepcopy(self.MESSAGES),
             button_classes=copy.deepcopy(self.BUTTON_CLASSES),
+            default_login_enabled=True,
         )
 
         """Creates and Auth object responsible for handling
@@ -213,9 +219,8 @@ class Auth(Fixture):
         self.use_username = use_username  # if False, uses email only
         self.use_phone_number = use_phone_number
         # The self._link variable is not thread safe (only intended for testing)
-        self.extra_auth_user_fields = extra_fields
+        self.extra_auth_user_fields = extra_fields or []
         self._link = None
-        self.extra_auth_user_fields = extra_fields
         if db and define_tables:
             self.define_tables()
         self.plugins = {}
@@ -309,7 +314,7 @@ class Auth(Fixture):
                         readable=False,
                     )
                 )
-            db.define_table("auth_user", *auth_fields, *self.extra_auth_user_fields)
+            db.define_table("auth_user", *(auth_fields + self.extra_auth_user_fields))
 
     @property
     def signature(self):
@@ -454,6 +459,11 @@ class Auth(Fixture):
 
     def login(self, email, password):
         db = self.db
+
+        # Default incase they are None or an error will occur.
+        email = "" if email is None else email
+        password = "" if password is None else password
+
         if "email_auth" in self.plugins:
             email = email.lower()
             if self.plugins["email_auth"].validate_credentials(email, password):
@@ -635,17 +645,44 @@ class Auth(Fixture):
         email = "%s@example.com" % token
         return db(db.auth_user.email == email).count() > 0
 
+    def get_or_delete_existing_unverified_account(self, email):
+        db = self.db
+        row = db(db.auth_user.email == email).select(limitby=(0,1)).first()
+        # if we have a user with this email and incomplete registration delete it
+        if row and row.action_token.startswith("pending-registration:"):
+            row.delete_record()
+            return None
+        return row 
+    
     def get_or_register_user(self, user):
         db = self.db
-        row = db(db.auth_user.sso_id == user["sso_id"]).select(limitby=(0, 1)).first()
-        data = user
-        if row:
-            if any(user[key] != row[key] for key in user):
-                row.update_record(**user)
-            data["id"] = row["id"]
+        # if the we have an email for the user
+        if 'email' in user:
+            # return a user if exists and has a verified email
+            row = self.get_or_delete_existing_unverified_account(user['email'])
+        # else retrieve the user from the sso_id
         else:
-            data["id"] = db.auth_user.insert(**db.auth_user._filter_fields(user))
-        return data
+            row = db(db.auth_user.sso_id == user["sso_id"]).select(limitby=(0, 1)).first()
+        # if we have found a candidate user
+        if row:
+            # we expect the email to match if provided
+            if 'email' in user and row.email != user['email']:
+                return None
+            # we can update all the other information provided by the SSO
+            if any(user[key] != row[key] for key in user if not key == 'username'):
+                row.update_record(**user)
+            user["id"] = row["id"]
+        # if we do not have a candidate user we need to create one
+        else:
+            # we expect an email to unable to create account
+            if not 'email' in user:
+                return None
+            # if we expect a username but not provided, user email as username
+            if self.use_username and 'username' not in user:
+                user['username'] = user['email']
+            # create the user
+            user["id"] = db.auth_user.insert(**db.auth_user._filter_fields(user))
+        return user
 
     # Private methods
 
@@ -828,6 +865,7 @@ class AuthAPI:
         "config",
         "register",
         "login",
+        "logout",
         "request_reset_password",
         "reset_password",
     ]
@@ -853,11 +891,16 @@ class AuthAPI:
     @staticmethod
     @api_wrapper
     def register(auth):
+        if request.json is None:
+            return auth._error("no json post payload")
+        auth.get_or_delete_existing_unverified_account(request.json.get('email'))
         return auth.register(request.json, send=True).as_dict()
 
     @staticmethod
     @api_wrapper
     def login(auth):
+        if request.json is None:
+            return auth._error("no json post payload")
         username, password = request.json.get("email"), request.json.get("password")
         if not all(isinstance(_, str) for _ in [username, password]):
             return auth._error("Invalid Credentials")
@@ -895,6 +938,8 @@ class AuthAPI:
     @staticmethod
     @api_wrapper
     def request_reset_password(auth):
+        if request.json is None:
+            return auth._error("no json post payload")
         if not auth.request_reset_password(**request.json):
             return auth._error("invalid user")
         return {}
@@ -903,6 +948,8 @@ class AuthAPI:
     @api_wrapper
     def reset_password(auth):
         # check the new_password2 only if passed
+        if request.json is None:
+            return auth._error("no json post payload")
         res = auth.reset_password(
             request.json.get("token"),
             request.json.get("new_password"),
@@ -928,6 +975,8 @@ class AuthAPI:
     @staticmethod
     @api_wrapper
     def change_password(auth):
+        if request.json is None:
+            return auth._error("no json post payload")
         return auth.change_password(
             auth.get_user(safe=False),  # refactor make faster
             request.json.get("new_password"),
@@ -937,6 +986,8 @@ class AuthAPI:
     @staticmethod
     @api_wrapper
     def change_email(auth):
+        if request.json is None:
+            return auth._error("no json post payload")
         return auth.change_email(
             auth.get_user(safe=False),
             request.json.get("new_email"),
@@ -948,6 +999,8 @@ class AuthAPI:
     def profile(auth):
         if request.method == "GET":
             return {"user": auth.get_user()}
+        if request.json is None:
+            return auth._error("no json post payload")
         else:
             return auth.update_profile(auth.get_user(), **request.json)
 
@@ -981,6 +1034,12 @@ class DefaultAuthForms:
                 )
                 break
         button_name = self.auth.param.messages["buttons"]["sign-up"]
+        # if the form is submitted, before any validation
+        # delete any unverified account with the same email
+        if request.method == 'POST':
+            email = request.forms.get('email')
+            if email:
+                self.auth.get_or_delete_existing_unverified_account(email)
         form = Form(fields, submit_value=button_name, formstyle=self.formstyle)
         user = None
         if form.accepted:
@@ -1012,7 +1071,30 @@ class DefaultAuthForms:
             )
         return form
 
+    def login_buttons(self):
+
+        top_buttons = []
+
+        for name, plugin in self.auth.plugins.items():
+            url = "../auth/plugin/" + name + "/login"
+            if request.query.get("next"):
+                url = url + "?next=" + request.query.get("next")
+            if (
+                name != "email_auth"
+            ):  #  do not add the top button for the email auth plugin
+                top_buttons.append(
+                    A(plugin.label + " Login", _href=url, _role="button")
+                )
+        return top_buttons
+
     def login(self):
+
+        top_buttons = self.login_buttons()
+
+        # if we do not allow we only display the plugin login buttons
+        if not self.auth.param.default_login_enabled:
+            return DIV(*top_buttons)
+
         fields = [
             Field(
                 "username",
@@ -1028,7 +1110,7 @@ class DefaultAuthForms:
         else:
             fields[0].label = self.auth.db.auth_user.email.label
         fields[1].label = self.auth.db.auth_user.password.label
-
+        
         button_name = self.auth.param.messages["buttons"]["sign-in"]
         form = Form(
             fields,
@@ -1043,16 +1125,10 @@ class DefaultAuthForms:
             )
             form.accepted = not error
             form.errors["username"] = error
-            if user:
-                self.auth.store_user_in_session(user["id"])
-                self._postprocessing("login", form, user)
-        top_buttons = []
-        for name, plugin in self.auth.plugins.items():
-            url = "../auth/plugin/" + name + "/login"
-            if self.auth.next["login"]:
-                url = url + "?next=" + self.auth.next["login"]
-            top_buttons.append(A(plugin.label + " Login", _href=url, _role="button"))
-
+        if user:
+            self.auth.store_user_in_session(user["id"])
+            self._postprocessing("login", form, user)
+            
         if self.auth.allows("register"):
             form.param.sidecar.append(
                 A(
