@@ -1,6 +1,7 @@
 import base64
 import copy
 from functools import reduce
+from urllib.parse import urlparse
 
 from yatl.helpers import (
     DIV,
@@ -78,6 +79,7 @@ class GridClassStyle:
         "grid-cell-type-text": "grid-cell-type-text",
         "grid-cell-type-boolean": "grid-cell-type-boolean",
         "grid-cell-type-float": "grid-cell-type-float",
+        "grid-cell-type-decimal": "grid-cell-type-decimal",
         "grid-cell-type-int": "grid-cell-type-int",
         "grid-cell-type-date": "grid-cell-type-date",
         "grid-cell-type-time": "grid-cell-type-time",
@@ -118,6 +120,7 @@ class GridClassStyle:
         "grid-cell-type-text": "vertical-align: middle; text-align: left; text-overflow: ellipsis; max-width: 200px",
         "grid-cell-type-boolean": "white-space: nowrap; vertical-align: middle; text-align: center",
         "grid-cell-type-float": "white-space: nowrap; vertical-align: middle; text-align: right",
+        "grid-cell-type-decimal": "white-space: nowrap; vertical-align: middle; text-align: right",
         "grid-cell-type-int": "white-space: nowrap; vertical-align: middle; text-align: right",
         "grid-cell-type-date": "white-space: nowrap; vertical-align: middle; text-align: right",
         "grid-cell-type-time": "white-space: nowrap; vertical-align: middle; text-align: right",
@@ -146,11 +149,11 @@ class GridClassStyleBulma(GridClassStyle):
 
     classes = {
         "grid-wrapper": "grid-wrapper field",
-        "grid-header": "grid-header pb-2",
+        "grid-header": "grid-header pb-2 is-clearfix",
         "grid-new-button": "grid-new-button button",
         "grid-search": "grid-search is-pulled-right pb-2",
-        "grid-table-wrapper": "grid-table-wrapper table_wrapper",
-        "grid-table": "grid-table table is-bordered is-striped is-hoverable is-fullwidth",
+        "grid-table-wrapper": "grid-table-wrapper table_wrapper table-container mb-1",
+        "grid-table": "grid-table table is-bordered is-striped is-hoverable is-fullwidth is-clearfix",
         "grid-sorter-icon-up": "grid-sort-icon-up fas fa-sort-up is-pulled-right",
         "grid-sorter-icon-down": "grid-sort-icon-down fas fa-sort-down is-pulled-right",
         "grid-th-action-button": "grid-col-action-button is-narrow",
@@ -162,7 +165,7 @@ class GridClassStyleBulma(GridClassStyle):
         "grid-details-button": "grid-details-button button is-small",
         "grid-edit-button": "grid-edit-button button is-small",
         "grid-delete-button": "grid-delete-button button is-small",
-        "grid-footer": "grid-footer",
+        "grid-footer": "grid-footer pb-4",
         "grid-info": "grid-info is-pulled-left",
         "grid-pagination": "grid-pagination is-pulled-right",
         "grid-pagination-button": "grid-pagination-button button is-small",
@@ -171,6 +174,7 @@ class GridClassStyleBulma(GridClassStyle):
         "grid-cell-type-text": "grid-cell-type-text",
         "grid-cell-type-boolean": "grid-cell-type-boolean has-text-centered",
         "grid-cell-type-float": "grid-cell-type-float",
+        "grid-cell-type-decimal": "grid-cell-type-decimal",
         "grid-cell-type-int": "grid-cell-type-int",
         "grid-cell-type-date": "grid-cell-type-date",
         "grid-cell-type-time": "grid-cell-type-time",
@@ -207,16 +211,19 @@ class GridClassStyleBulma(GridClassStyle):
         "grid-pagination": "",
         "grid-pagination-button": "margin-left: .25em;",
         "grid-pagination-button-current": "margin-left: .25em;",
-        "grid-cell-type-string": "",
-        "grid-cell-type-text": "",
-        "grid-cell-type-boolean": "",
-        "grid-cell-type-float": "",
-        "grid-cell-type-int": "",
-        "grid-cell-type-date": "",
-        "grid-cell-type-time": "",
-        "grid-cell-type-datetime": "",
-        "grid-cell-type-upload": "",
-        "grid-cell-type-list": "",
+        "grid-cell-type-string": "vertical-align: top; text-align: left; text-overflow: ellipsis;",
+        "grid-cell-type-text": "vertical-align: top; text-align: left; text-overflow: ellipsis;",
+        "grid-cell-type-boolean": "vertical-align: top; text-align: center",
+        "grid-cell-type-float": "vertical-align: top; text-align: right",
+        "grid-cell-type-decimal": "vertical-align: top; text-align: right",
+        "grid-cell-type-int": "vertical-align: top; text-align: center;",
+        "grid-cell-type-integer": "vertical-align: top; text-align: center;",
+        "grid-cell-type-reference": "vertical-align: top; text-align: center;",
+        "grid-cell-type-date": "vertical-align: top; text-align: center",
+        "grid-cell-type-time": "vertical-align: top; text-align: center",
+        "grid-cell-type-datetime": "vertical-align: top; text-align: center",
+        "grid-cell-type-upload": "vertical-align: top; text-align: center",
+        "grid-cell-type-list": "vertical-align: top; text-align: left",
         # specific for custom form
         "search_form": "",
         "search_form_table": "",
@@ -236,7 +243,7 @@ class Grid:
             "<script>document.write((new Date(%s,%s,%s,%s,%s,%s)).toLocaleString())</script>"
             % (
                 value.year,
-                value.month,
+                value.month - 1,
                 value.day,
                 value.hour,
                 value.minute,
@@ -255,7 +262,7 @@ class Grid:
             '<script>document.write((new Date(%s,%s,%s)).toLocaleString().split(",")[0])</script>'
             % (
                 value.year,
-                value.month,
+                value.month - 1,
                 value.day,
             )
         )
@@ -273,6 +280,7 @@ class Grid:
         search_form=None,
         search_queries=None,
         fields=None,
+        field_id=None,
         show_id=False,
         orderby=None,
         left=None,
@@ -299,6 +307,8 @@ class Grid:
         :param search_queries: future use - pass a dict of name and a search query
         :param fields: list of fields to display on the list page, if blank, glean tablename from first query
         :              and use all fields of that table
+        :param field_id: the id field of the primary table for the grid - if there are multiple tables (joined)
+                            then the table for this field is used as the table for edit/details/delete
         :param show_id: show the record id field on list page - default = False
         :param orderby: pydal orderby field or list of fields
         :param left: if joining other tables, specify the pydal left expression here
@@ -321,6 +331,7 @@ class Grid:
         self.param = Param(
             query=query,
             fields=fields,
+            field_id=field_id,
             show_id=show_id,
             orderby=orderby,
             left=left,
@@ -370,6 +381,7 @@ class Grid:
         self.use_tablename = self.is_join()
         self.formatters = {}
         self.formatters_by_type = copy.copy(Grid.FORMATTERS_BY_TYPE)
+        self.attributes_plugin = AttributesPlugin(request)
 
         if auto_process:
             self.process()
@@ -394,7 +406,10 @@ class Grid:
 
         parts = self.path.split("/")
         self.action = parts[0] or "select"
-        self.tablename = self.get_tablenames(self.param.query)[0]  # what if there ar 2?
+        if self.param.field_id:
+            self.tablename = str(self.param.field_id._table)
+        else:
+            self.tablename = self.get_tablenames(self.param.query)[0]
         self.record_id = safe_int(parts[1] if len(parts) > 1 else None, default=None)
 
         if self.param.fields:
@@ -435,11 +450,13 @@ class Grid:
                         db[self.tablename][field.name].readable = False
                         db[self.tablename][field.name].writable = False
 
+            attrs = self.attributes_plugin.form(url=request.url)
             self.form = Form(
                 db[self.tablename],
                 record=self.record_id,
                 readonly=readonly,
                 formstyle=self.param.formstyle,
+                **attrs,
             )
             if self.action == "new":
                 if self.param.new_sidecar:
@@ -473,6 +490,10 @@ class Grid:
 
         elif self.action == "delete":
             db(db[self.tablename].id == self.record_id).delete()
+
+            url = parse_referer(request)
+            if url and url.query:
+                self.endpoint += "?%s" % url.query
             redirect(self.endpoint)
 
         elif self.action == "select":
@@ -485,7 +506,8 @@ class Grid:
             key_is_missing = True
             for field in self.param.fields:
                 if (
-                    field.table._tablename == pt._tablename
+                    not isinstance(field, FieldVirtual)
+                    and field.table._tablename == pt._tablename
                     and field.name == pt._id.name
                 ):
                     key_is_missing = False
@@ -539,7 +561,10 @@ class Grid:
                 self.page_end = self.total_number_of_rows
 
             if self.param.fields:
-                self.rows = db(query).select(*self.param.fields, **select_params)
+                param_fields = [
+                    x for x in self.param.fields if not isinstance(x, FieldVirtual)
+                ]  # don't send virtual fields to DAL select - will calculate them later
+                self.rows = db(query).select(*param_fields, **select_params)
             else:
                 self.rows = db(query).select(**select_params)
 
@@ -591,7 +616,7 @@ class Grid:
         row_id=None,
         name="grid-button",
         row=None,
-        **attr,
+        **attrs,
     ):
         separator = "?"
         if row_id:
@@ -605,6 +630,18 @@ class Grid:
                 " ".join(items) if isinstance(items, (list, tuple)) else " %s" % items
             )
 
+        if callable(additional_classes):
+            additional_classes = additional_classes(row)
+
+        if callable(additional_styles):
+            additional_styles = additional_styles(row)
+
+        if callable(override_classes):
+            override_classes = override_classes(row)
+
+        if callable(override_styles):
+            override_styles = override_styles(row)
+
         if override_classes:
             classes = join(override_classes)
         elif additional_classes:
@@ -617,15 +654,15 @@ class Grid:
         if callable(url):
             url = url(row)
 
+        attrs.update(self.attributes_plugin.link(url=url))
         link = A(
             I(_class="fa %s" % icon),
-            _href=url,
             _role="button",
             _class=classes,
             _message=message,
             _title=button_text,
             _style=styles,
-            **attr,
+            **attrs,
         )
         if self.param.include_action_button_text:
             link.append(
@@ -638,7 +675,6 @@ class Grid:
         return link
 
     def render_default_form(self):
-
         search_type = safe_int(request.query.get("search_type", 0), default=0)
         search_string = request.query.get("search_string")
         options = [
@@ -650,7 +686,8 @@ class Grid:
             for key in request.query
             if not key in ("search_type", "search_string")
         ]
-        form = FORM(*hidden_fields, **dict(_method="GET", _action=self.endpoint))
+        attrs = self.attributes_plugin.form(url=self.endpoint)
+        form = FORM(*hidden_fields, **attrs)
         select = SELECT(
             *options,
             **dict(
@@ -747,19 +784,24 @@ class Grid:
                 #  add the sort order query parm
                 sort_query_parms = dict(self.query_parms)
 
-                if key == sort_order:
+                attrs = {}
+                if isinstance(field, FieldVirtual):
+                    col = DIV(heading)
+                elif key == sort_order:
                     sort_query_parms["orderby"] = "~" + key
-                    href = URL(self.endpoint, vars=sort_query_parms)
-                    col = A(heading, up, _href=href)
+                    url = URL(self.endpoint, vars=sort_query_parms)
+                    attrs = self.attributes_plugin.link(url=url)
+                    col = A(heading, up, **attrs)
                 else:
                     sort_query_parms["orderby"] = key
-                    href = URL(self.endpoint, vars=sort_query_parms)
-                    col = A(heading, dw if "~" + key == sort_order else "", _href=href)
+                    url = URL(self.endpoint, vars=sort_query_parms)
+                    attrs = self.attributes_plugin.link(url=url)
+                    col = A(heading, dw if "~" + key == sort_order else "", **attrs)
                 columns.append((key, col))
 
         thead = THEAD(_class=self.param.grid_class_style.classes.get("grid-thead", ""))
         for key, col in columns:
-            col_class = "grid-col-%s" % key
+            col_class = " grid-col-%s" % key
             thead.append(
                 TH(
                     col,
@@ -769,7 +811,13 @@ class Grid:
                 )
             )
 
-        if self.param.details or self.param.editable or self.param.deletable:
+        if (
+            self.param.details
+            or self.param.editable
+            or self.param.deletable
+            or self.param.pre_action_buttons
+            or self.param.post_action_buttons
+        ):
             thead.append(
                 TH("", **self.param.grid_class_style.get("grid-th-action-button"))
             )
@@ -789,7 +837,13 @@ class Grid:
         :param field:
         :return:
         """
-        if self.use_tablename:
+        if isinstance(field, FieldVirtual):
+            #  handle virtual fields in table display
+            if self.use_tablename:
+                field_value = field.f(row[field.tablename])
+            else:
+                field_value = field.f(row)
+        elif self.use_tablename:
             field_value = row[field.tablename][field.name]
         else:
             field_value = row[field.name]
@@ -801,9 +855,11 @@ class Grid:
         )
 
         class_type = "grid-cell-type-%s" % str(field.type).split(":")[0].split("(")[0]
-        class_col = "grid-col-%s" % key
+        class_col = " grid-col-%s" % key
         td = TD(
-            formatter(field_value),
+            formatter(field_value)
+            if formatter.__code__.co_argcount == 1  # if formatter has only 1 argument
+            else formatter(field_value, row),
             _class=(
                 self.param.grid_class_style.classes.get("grid-td", "")
                 + " "
@@ -830,7 +886,7 @@ class Grid:
         tbody = TBODY()
         for row in self.rows:
             #  find the row id - there may be nested tables....
-            if self.use_tablename and self.tablename in row:
+            if self.use_tablename and self.tablename in row and "id" not in row:
                 row_id = row[self.tablename]["id"]
             else:
                 row_id = row["id"]
@@ -853,6 +909,7 @@ class Grid:
                 (self.param.details and self.param.details != "")
                 or (self.param.editable and self.param.editable != "")
                 or (self.param.deletable and self.param.deletable != "")
+                or (self.param.post_action_buttons or self.param.pre_action_buttons)
             ):
                 classes = (
                     self.param.grid_class_style.classes.get("grid-td", "")
@@ -867,6 +924,14 @@ class Grid:
                 td = TD(_class=classes, _style=styles)
                 if self.param.pre_action_buttons:
                     for btn in self.param.pre_action_buttons:
+                        if callable(btn):
+                            # a button can be a callable, to indicate whether or not a button should
+                            # be displayed. call the function with the row object
+                            btn = btn(row)
+                            if btn is None:
+                                # if None was returned, no button is available for this row: ignore this value in the
+                                # list
+                                continue
                         if btn.onclick:
                             btn.url = None
                         td.append(
@@ -889,9 +954,9 @@ class Grid:
                     details_url += "/%s?%s" % (row_id, self.referrer)
                     td.append(
                         self.render_action_button(
-                            details_url,
-                            self.param.details_action_button_text,
-                            "fa-id-card",
+                            url=details_url,
+                            button_text=self.param.details_action_button_text,
+                            icon="fa-id-card",
                             name="grid-details-button",
                         )
                     )
@@ -904,9 +969,9 @@ class Grid:
                     edit_url += "/%s?%s" % (row_id, self.referrer)
                     td.append(
                         self.render_action_button(
-                            edit_url,
-                            self.param.edit_action_button_text,
-                            "fa-edit",
+                            url=edit_url,
+                            button_text=self.param.edit_action_button_text,
+                            icon="fa-edit",
                             name="grid-edit-button",
                         )
                     )
@@ -917,19 +982,30 @@ class Grid:
                     else:
                         delete_url = self.endpoint + "/delete"
                     delete_url += "/%s?%s" % (row_id, self.referrer)
+                    attrs = self.attributes_plugin.confirm(
+                        message="Are you sure you want to delete?"
+                    )
                     td.append(
                         self.render_action_button(
-                            delete_url,
-                            self.param.delete_action_button_text,
-                            "fa-trash",
+                            url=delete_url,
+                            button_text=self.param.delete_action_button_text,
+                            icon="fa-trash",
                             additional_classes="confirmation",
                             message="Delete record",
                             name="grid-delete-button",
-                            _onclick="if(!confirm('sure you want to delete')) return false;",
+                            **attrs,
                         )
                     )
                 if self.param.post_action_buttons:
                     for btn in self.param.post_action_buttons:
+                        if callable(btn):
+                            # a button can be a callable, to indicate whether or not a button should
+                            # be displayed. call the function with the row object
+                            btn = btn(row)
+                            if btn is None:
+                                # if None was returned, no button is available for this row: ignore this value in the
+                                # list
+                                continue
                         if btn.onclick:
                             btn.url = None
                         td.append(
@@ -966,12 +1042,15 @@ class Grid:
                 if is_current
                 else "grid-pagination-button"
             )
+            attrs = self.attributes_plugin.link(
+                url=URL(self.endpoint, vars=pager_query_parms)
+            )
             pager.append(
                 A(
                     page_number,
                     **self.param.grid_class_style.get(page_name),
                     _role="button",
-                    _href=URL(self.endpoint, vars=pager_query_parms),
+                    **attrs,
                 )
             )
             previous_page_number = page_number
@@ -988,6 +1067,8 @@ class Grid:
             else:
                 create_url = self.endpoint + "/new"
 
+            create_url += "?%s" % self.referrer
+
             grid_header.append(
                 self.render_action_button(
                     create_url,
@@ -1033,7 +1114,7 @@ class Grid:
                 else self.total_number_of_rows,
                 self.total_number_of_rows,
             )
-        )
+        ) if self.number_of_pages > 0 else row_count.append("No rows to display")
         footer.append(row_count)
 
         #  build the pager
@@ -1090,3 +1171,110 @@ class Grid:
             else:
                 items += [self.param.left]
         return len(self.get_tablenames(*items)) > 1
+
+
+def parse_referer(r):
+    """
+    Get the referrer from the request query and use urlparse to parse it into a url object
+
+    :param r: The request object to be interrogated
+
+    :return: the URL object or None if no URL object obtained/decoded
+    """
+    referrer = r.query.get("_referrer")
+    url = None
+    if referrer:
+        url_string = base64.b16decode(referrer.encode("utf8")).decode("utf8")
+        if url_string:
+            url = urlparse(url_string)
+
+    return url
+
+
+def get_parent(path, parent_field):
+    """
+    try to find the parent id for a parent/child table relationship
+
+    :param path: the path var from the method calling the grid
+    :param child_table: the child table of the parent/child relationship ex. db.child_table
+    :param parent_field_name: this is the name of the field in the child table that points to the parent table
+                              this is a string value
+
+    :return parent_id: the id of the parent record
+    """
+    parent_id = request.query.get("parent_id")
+
+    child_table = parent_field._table
+    fn = parent_field.name
+
+    #  find the record id of the parent from the child table record
+    if path:
+        parts = path.split("/")
+        record_id = parts[1] if len(parts) > 1 else None
+        if record_id:
+            r = child_table(record_id)
+            if r:
+                parent_id = r[fn]
+
+    #  not passed in, check in the form
+    if not parent_id:
+        parent_id = request.forms.get(fn)
+
+    #  not found yet, check in the referer
+    if not parent_id:
+        referrer = request.query.get("_referrer")
+        if referrer:
+            kvp = base64.b16decode(referrer.encode("utf8")).decode("utf8")
+            if "parent_id" in kvp:
+                parent_id = kvp.split("parent_id=")[1]
+
+    try:
+        parent_id = parent_id.split("&")[0]
+    except:
+        pass
+
+    return parent_id
+
+
+class AttributesPlugin:
+    def __init__(self, target_element=None):
+        self.target_element = target_element
+        self.default_attrs = {}
+
+    def form(self, url):
+        attrs = copy.copy(self.default_attrs)
+        return attrs
+
+    def link(self, url):
+        attrs = copy.copy(self.default_attrs)
+        attrs["_href"] = url
+        return attrs
+
+    def confirm(self, message):
+        attrs = copy.copy(self.default_attrs)
+        attrs["_onclick"] = "if(!confirm('%s')) return false;" % message
+        return attrs
+
+
+class AttributesPluginHtmx(AttributesPlugin):
+    def __init__(self, target_element):
+        super().__init__(target_element)
+        self.default_attrs = {
+            "_hx-target": self.target_element,
+            "_hx-swap": "innerHTML",
+        }
+
+    def form(self, url):
+        attrs = copy.copy(self.default_attrs)
+        attrs["_hx-post"] = url
+        return attrs
+
+    def link(self, url):
+        attrs = copy.copy(self.default_attrs)
+        attrs["_hx-get"] = url
+        return attrs
+
+    def confirm(self, message):
+        attrs = copy.copy(self.default_attrs)
+        attrs["_hx-confirm"] = message
+        return attrs
